"""Module with the Hermes audio player class."""
import io
import json
import wave

from humanfriendly import format_size
import pyaudio

from hermes_audio_server.mqtt import MQTTClient

PLAY_BYTES = 'hermes/audioServer/{}/playBytes/+'
PLAY_FINISHED = 'hermes/audioServer/{}/playFinished'
CHUNK = 256


class AudioPlayer(MQTTClient):
    """This class creates an MQTT client that acts as an audio player for the
    Hermes protocol.
    """

    def initialize(self):
        """Initialize a Hermes audio player."""
        self.audio = pyaudio.PyAudio()
        self.audio_out = self.audio.get_default_output_device_info()['name']
        print('Connected to audio output {}.'.format(self.audio_out))

    def on_connect(self, client, userdata, flags, result_code):
        """Callback that is called when the audio player connects to the MQTT
        broker."""
        super(AudioPlayer, self).on_connect(client, userdata, flags, result_code)
        play_bytes = PLAY_BYTES.format(self.config.site)
        self.mqtt.subscribe(play_bytes)
        self.mqtt.message_callback_add(play_bytes, self.on_play_bytes)
        print('Subscribed to {}.'.format(play_bytes))

    def on_play_bytes(self, client, userdata, message):
        """Callback that is called when the audio player receives a PLAY_BYTES
        message on MQTT.
        """
        request_id = message.topic.split('/')[4]
        length = format_size(len(message.payload), binary=True)
        print('Received an audio message of length {}'
              ' with request id {} on site {}.'.format(length,
                                                       request_id,
                                                       self.config.site))

        with io.BytesIO(message.payload) as wav_buffer:
            try:
                with wave.open(wav_buffer, 'rb') as wav:
                    sample_width = wav.getsampwidth()
                    sample_format = self.audio.get_format_from_width(sample_width)
                    n_channels = wav.getnchannels()
                    frame_rate = wav.getframerate()

                    if self.verbose:
                        print('Sample width: {}'.format(sample_width))
                        print('Channels: {}'.format(n_channels))
                        print('Frame rate: {}'.format(frame_rate))

                    stream = self.audio.open(format=sample_format,
                                             channels=n_channels,
                                             rate=frame_rate,
                                             output=True)

                    data = wav.readframes(CHUNK)

                    while len(data) > 0:
                        stream.write(data)
                        data = wav.readframes(CHUNK)

                    stream.stop_stream()
                    stream.close()

                    self.mqtt.publish(PLAY_FINISHED.format(self.config.site),
                                      json.dumps({'id': request_id,
                                                  'siteId': self.config.site}))
                    print('Finished playing audio message with id {}'
                          ' on device {} on site {}'.format(request_id,
                                                            self.audio_out,
                                                            self.config.site))
            except wave.Error as error:
                print('Error: {}'.format(str(error)))
            except EOFError:
                print('Error: end of WAV buffer')



"""Module with an MQTT client. Both the audio player and audio recorder class
inherit from this class.
"""
from paho.mqtt.client import Client


class MQTTClient:
    """This class represents an MQTT client for Hermes Audio Server.

    This is an abstract base class. You don't instantiate an object of this
    class, but an object of one of its subclasses.
    """

    def __init__(self, config, verbose):
        """Initialize an MQTT client.

        Args:
            config (:class:`.ServerConfig`): The configuration of
                the MQTT client.
            verbose (bool): Whether or not the MQTT client runs in verbose
                mode.
        """
        self.config = config
        self.verbose = verbose
        self.mqtt = Client()

        self.initialize()

        self.mqtt.on_connect = self.on_connect
        self.mqtt.on_disconnect = self.on_disconnect
        self.connect()

    def connect(self):
        """Connect to the MQTT broker defined in the configuration."""
        # Set up MQTT authentication.
        if self.config.mqtt.auth.enabled:
            self.mqtt.username_pw_set(self.config.mqtt.auth.username,
                                      self.config.mqtt.auth.password)

        # Set up an MQTT TLS connection.
        if self.config.mqtt.tls.enabled:
            self.mqtt.tls_set(ca_certs=self.config.mqtt.tls.ca_certs,
                              certfile=self.config.mqtt.tls.client_cert,
                              keyfile=self.config.mqtt.tls.client_key)

        self.mqtt.connect(self.config.mqtt.host, self.config.mqtt.port)

    def initialize(self):
        """Initialize the MQTT client."""

    def start(self):
        """Start the event loop to the MQTT broker so the audio server starts
        listening to MQTT topics and the callback methods are called.
        """
        self.mqtt.loop_forever()

    def on_connect(self, client, userdata, flags, result_code):
        """Callback that is called when the client connects to the MQTT broker.
        """
        print('Connected to MQTT broker {}:{}'
              ' with result code {}.'.format(self.config.mqtt.host,
                                             self.config.mqtt.port,
                                             result_code))

    def on_disconnect(self, client, userdata, flags, result_code):
        """Callback that is called when the client connects from the MQTT
        broker."""
        print('Disconnected with result code {}.'.format(result_code))



"""Module with the Hermes audio recorder class."""
import io
import json
from threading import Thread
import wave

import pyaudio
import webrtcvad

from hermes_audio_server.mqtt import MQTTClient

AUDIO_FRAME = 'hermes/audioServer/{}/audioFrame'
CHANNELS = 1
CHUNK = 320  # = FRAME_RATE * 20 / 1000 (20 ms)
FRAME_RATE = 16000
SAMPLE_WIDTH = 2

VAD_DOWN = 'hermes/voiceActivity/{}/vadDown'
VAD_UP = 'hermes/voiceActivity/{}/vadUp'


# TODO: Call stream.stop_stream() and stream.close()
class AudioRecorder(MQTTClient):
    """This class creates an MQTT client that acts as an audio recorder for the
    Hermes protocol.
    """

    def initialize(self):
        """Initialize a Hermes audio recorder."""
        self.audio = pyaudio.PyAudio()
        self.audio_in = self.audio.get_default_input_device_info()['name']
        print('Connected to audio input {}.'.format(self.audio_in))

        if self.config.vad.enabled:
            print('Voice Activity Detection enabled with mode {}'.format(self.config.vad.mode))
            self.vad = webrtcvad.Vad(self.config.vad.mode)

    def start(self):
        """Start the event loop to the MQTT broker and start the audio
        recording."""
        Thread(target=self.send_audio_frames, daemon=True).start()
        super(AudioRecorder, self).start()

    def publish_frames(self, frames):
        """Publish frames on MQTT."""
        with io.BytesIO() as wav_buffer:
            with wave.open(wav_buffer, 'wb') as wav:
                # pylint: disable=no-member
                wav.setnchannels(CHANNELS)
                wav.setsampwidth(SAMPLE_WIDTH)
                wav.setframerate(FRAME_RATE)
                wav.writeframes(frames)
            self.mqtt.publish(AUDIO_FRAME.format(self.config.site),
                              wav_buffer.getvalue())

    def publish_vad_status_message(self, message):
        """Publish a status message about the VAD on MQTT."""
        if self.config.vad.status_messages:
            self.mqtt.publish(message.format(self.config.site),
                              json.dumps({'siteId': self.config.site,
                                          'signalMs': 0}))  # Not used

    def send_audio_frames(self):
        """Send the recorded audio frames continuously in AUDIO_FRAME
        messages on MQTT.
        """
        stream = self.audio.open(format=pyaudio.paInt16, channels=CHANNELS,
                                 rate=FRAME_RATE, input=True,
                                 frames_per_buffer=CHUNK)

        print('Started broadcasting audio from device {}'
              ' on site {}'.format(self.audio_in, self.config.site))

        in_speech = False
        silence_frames = int(FRAME_RATE / CHUNK * self.config.vad.silence)

        # TODO: Simplify if ... if ...
        while True:
            frames = stream.read(CHUNK, exception_on_overflow=False)
            if self.config.vad.enabled and self.vad.is_speech(frames, FRAME_RATE):
                if not in_speech:
                    in_speech = True
                    silence_frames = int(FRAME_RATE / CHUNK * self.config.vad.silence)
                    print('Voice activity started on site {}'.format(self.config.site))
                    self.publish_vad_status_message(VAD_UP)
                self.publish_frames(frames)
            elif self.config.vad.enabled:
                if in_speech and silence_frames > 0:
                    self.publish_frames(frames)
                    silence_frames -= 1
                elif in_speech:
                    in_speech = False
                    print('Voice activity stopped on site {}'.format(self.config.site))
                    self.publish_vad_status_message(VAD_DOWN)
            else:
                self.publish_frames(frames)



from importlib import import_module
from pathlib import Path
from setuptools import setup, find_packages

SRC_ROOT = 'src'
BIN_ROOT = 'bin/'

about = import_module(SRC_ROOT + '.hermes_audio_server.about')

with Path('README.md').open('r') as fh:
    long_description = fh.read()

with Path('requirements.txt').open('r') as fh:
    requirements = fh.read().splitlines()
    requirements = [requirement for requirement in requirements
                    if not requirement.startswith('#')]

binaries = [BIN_ROOT + about.PLAYER, BIN_ROOT + about.RECORDER]

setup(
    name=about.PROJECT,
    version=about.VERSION,
    description=about.DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    license=about.LICENSE,
    author=about.AUTHOR,
    author_email=about.EMAIL,
    url=about.GITHUB_URL,
    project_urls={
        'Documentation': about.DOC_URL,
        'Source': about.GITHUB_URL,
        'Tracker': about.TRACKER_URL,
    },
    packages=find_packages(SRC_ROOT),
    package_dir={'': SRC_ROOT},
    install_requires=requirements,
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Home Automation',
        'Topic :: Multimedia :: Sound/Audio :: Capture/Recording',
        'Topic :: Multimedia :: Sound/Audio :: Players'
    ],
    keywords=about.KEYWORDS,
    scripts=binaries)
